import { Request, Response } from "express";
import Joi from "joi";
import { Location } from "../db";
import sequelize from "sequelize";
import { pagination } from "../utils/pagination";

const LocationParamsSchema = Joi.object({
  id: Joi.string()
    .uuid({ version: 'uuidv4', separator: '-' })
    .required()
});

/**
 * Handles the GET /l/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function handleLocation(req: Request, res: Response) {
  try {
    const { id } = await LocationParamsSchema.validateAsync(req.params);
    const location = await Location.findOne({ where: { id } });
    if (location == null)
      throw Error("invalid location ID");

    // Find questions that are valid for today at the specified location.
    const now = new Date();
    const questions = await location.getQuestions({
      where: {
        validDateStart: { [sequelize.Op.lte]: now },
        validDateEnd: { [sequelize.Op.gte]: now }
      }
    }) || [];

    switch (questions.length) {
      case 0: // no valid question for today
        res.render("locationError");
        break;
      default: // too many valid questions... that is not good
        console.warn(`Too many valid questions at location ${location.id} (found ${questions.length} questions)`);
      case 1: // there is a question for today!
        const question = questions[0];
        res.redirect(`/q/${question.id}`);
        break;
    }
  } catch (err) {
    res.status(404).render("internalError", { msg: err });
    return;
  }
}

/**
 * Handles the GET /l path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function showLocations(req: Request, res: Response) {
  const locationsPerPage = 5;
  const locationCount = await Location.count();
  const maxPageCount = Math.ceil(locationCount / locationsPerPage);

  const page = Math.min(maxPageCount, Math.max(1, Number.parseInt((req.query.p || "1") as string)));

  const locations = await Promise.all((await Location.findAll({
    offset: (page - 1) * locationsPerPage,
    limit: locationsPerPage
  })).map(async location => {
    const frontendLocation = {
      id: location.id,
      userName: location.userName,
      questions: (await location.getQuestions()).map((question) => {
        const frontendQuestion = {
          id: question.id,
          text: question.text,
          dateBegin: question.validDateStart,
          dateEnd: question.validDateEnd,
          allowMultipleChoices: question.allowMultipleChoices
        };
        return frontendQuestion;
      }) || []
    };
    return frontendLocation;
  }));

  res.render("locations", { locations, pagination: pagination(page, locationCount) });
}
