import { Request, Response } from "express";
import Joi from "joi";
import { Answer, Choice, Question } from "../db";
import sequelize from "sequelize";

function shuffle(array: Array<Choice>) {
  // See https://stackoverflow.com/questions/2450954/how-to-randomize-shuffle-a-javascript-array

  let currentIndex = array.length;

  // While there remain elements to shuffle...
  while (currentIndex != 0) {

    // Pick a remaining element...
    let randomIndex = Math.floor(Math.random() * currentIndex);
    currentIndex--;

    // And swap it with the current element.
    [array[currentIndex], array[randomIndex]] = [
      array[randomIndex], array[currentIndex]];
  }
}

async function findQuestionValidNow(id: string, isAdmin: boolean) {
  if (isAdmin) {
    // Do not check time if we are admin.
    return await Question.findOne({
      where: { id }
    });
  } else {
    const now = new Date();
    return await Question.findOne({
      where: {
        id, validDateStart: { [sequelize.Op.lte]: now },
        validDateEnd: { [sequelize.Op.gte]: now }
      }
    });
  }
}

const QuestionParamsSchema = Joi.object({
  id: Joi.string()
    .uuid({ version: 'uuidv4', separator: '-' })
    .required()
});

function isAdmin(req: Request): boolean {
  if (req.session && req.session?.user)
    return req.session.user.admin as boolean;

  return false;
}

/**
 * Handles the GET /q/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function renderQuestion(req: Request, res: Response) {
  try {
    const { id } = await QuestionParamsSchema.validateAsync(req.params);
    const question = await findQuestionValidNow(id, isAdmin(req));
    if (question == null)
      throw Error("invalid question ID or not valid now");

    const choices = await question.getChoices();
    shuffle(choices);

    res.render("question", {
      question,
      choices: choices
    })
  } catch (err) {
    res.status(404).render("internalError", { msg: (err as Error).message });
    return;
  }
}

async function isAlreadyAnswered(questionId: string, playerId: string) {
  const answer = await Answer.findOne({ where: { playerId, questionId } });
  return answer != null;
}

function computeSolutions(choices: Choice[], selectedIds: string[]) {
  let solutions = [];

  let validChoices = 0;
  let invalidChoices = 0;
  let totalValidChoices = 0;
  for (let choice of choices) {
    let foundInUserSelectedChoices = selectedIds.findIndex((id) => 
      id == ((choice.id as unknown) as string)
    ) >= 0;

    if (choice.isAnswer) {
      totalValidChoices += 1;
      if (foundInUserSelectedChoices)
        validChoices += 1;
    } else if (foundInUserSelectedChoices) {
      invalidChoices += 1;
    }

    solutions.push({
      id: choice.id,
      isSelected: foundInUserSelectedChoices,
      isAnswer: choice.isAnswer,
      text: choice.text
    });
  }

  const isCorrect = validChoices === totalValidChoices && invalidChoices === 0;
  return {
    solutions,
    isCorrect
  };
}

/**
 * Handles the POST /q/:id path.
 * 
 * @param req The Express request object.
 * @param res The Express response object.
 */
export async function answerQuestion(req: Request, res: Response) {
  try {
    const questionId = (await QuestionParamsSchema.validateAsync(req.params)).id;
    const playerId = req.session?.user.id;

    if (await isAlreadyAnswered(questionId, playerId)) {
      res.status(400).render('alreadyAnswered');
      return;
    }

    const question = await findQuestionValidNow(questionId, isAdmin(req));
    if (question == null)
      throw Error("invalid question ID or not valid now");


    const userSelectedChoices: [] = Array.isArray(req.body.choices) ? req.body.choices : [req.body.choices];
    const choices = await question.getChoices();

    const { solutions, isCorrect } = computeSolutions(choices, userSelectedChoices);

    // Create Answer async.
    Answer.create({
      questionId,
      playerId,
      selectedChoices: userSelectedChoices.toString(),
      isCorrect
    });

    res.render("question", {
      isCorrect,
      question: question,
      choices: solutions,
      showSolution: true
    });
  } catch (err) {
    return res.status(404).render('internalError', { msg: (err as Error).message });
  }
}
